/*
 * KSR[T] Presents:
 *
 * pingsting v1.0
 *
 * ICMP ECHO Request client identifier.
 *
 * Authors:
 *
 * Anthony Osborne <ao@ksrt.org>
 * David Goldsmith <dhg@ksrt.org>
 *
 * Pingsting Web Site
 * http://www.ksrt.org/psting
 *
 * KSR[T] Main Web Site
 * http://www.ksrt.org
 *
 * 10/01/1999
 */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <pcap.h>

#include "ip.h"
#include "pingdb.h"

int  linklayer	= 0;
int  verbose	= 0;
int  promisc	= 0;
int  tstamp	= 1;
int  hostnames	= 1;
char ip_sname[MAXHOSTNAMELEN+1];

/*
 * Network Byte Order to Ascii. (avoiding struct in_addr).
 */
char *inet_nbotoa(u_long ipaddr_nbo)
{
        static char b[18];
        register char *p;

        memset(b, 0x0, sizeof(b));

        p = (char *)&ipaddr_nbo;
#define UC(b)   (((int)b)&0xff)
        (void)snprintf(b, sizeof(b),
            "%d.%d.%d.%d", UC(p[0]), UC(p[1]), UC(p[2]), UC(p[3]));
        return (b);
}

void print_echo(char *icmp, int datalen)
{
    icmph_t *icmph = (icmph_t *)icmp;
    char *icmpdata = icmp + sizeof(icmph_t);
    int i;

    if (icmph == NULL || icmpdata == NULL || datalen < 0)
	return;

    printf("icmp_type:\t%d\t",		icmph->icmp_type);
    printf("icmp_code:\t%d\t",		icmph->icmp_code);
    printf("icmp_cksum:   0x%04x\n",	icmph->icmp_cksum);
    printf("echo_id:\t%d\t",      	icmph->echo_id);
    printf("echo_seqno:\t%d\t",   	icmph->echo_seqno);
    printf("data_length:  %d\n", 	datalen);

    if (datalen == 0)
	return;

    printf("\"");
    for (i = 1 ; datalen > 0; datalen--, icmpdata++, i++) {
	printf("\\x%02x", *icmpdata & 0xFF);
	if (datalen == 1)
	    printf("\"");
	else if ((i % 16) == 0)
	    printf("\"\n\"");
    }
    printf("\n\n");
    fflush( stdout );
}

u_short in_cksum(u_short *addr, int len)
{
        register int nleft = len;
        register u_short *w = addr;
        register int sum = 0;
        u_short answer = 0;

        /*
         * Our algorithm is simple, using a 32 bit accumulator (sum), we add
         * sequential 16 bit words to it, and at the end, fold back all the
         * carry bits from the top 16 bits into the lower 16 bits.
         */
        while (nleft > 1)  {
                sum += *w++;
                nleft -= 2;
        }

        /* mop up an odd byte, if necessary */
        if (nleft == 1) {
                *(u_char *)(&answer) = *(u_char *)w ;
                sum += answer;
        }

        /* add back carry outs from top 16 bits to low 16 bits */
        sum = (sum >> 16) + (sum & 0xffff);     /* add hi 16 to low 16 */
        sum += (sum >> 16);                     /* add carry */
        answer = ~sum;                          /* truncate to 16 bits */
        return(answer);
}

/*
 * Default filter ensures IPv4 datagrams, no IP options and a length
 * greater than 28 although less than 1500 bytes.
 */
void ping_ident(u_char *u, const struct pcap_pkthdr *h, const u_char *p)
{
    iph_t    *iph;
    icmph_t  *icmph;
    pingdb_t *e = (pingdb_t *)pingdb;
    char     *dataoffset, **namep;
    char     ip_dname[MAXHOSTNAMELEN+1];
    int      datalen = 0, identified = 0;
    struct   timeval tp;
    struct   tm *tm;
    struct   hostent *hps, *hpd;
    struct   in_addr ia;

    p += linklayer;
    iph = (iph_t *)p;

    if (iph == NULL || h->caplen < sizeof(iph_t))
	return;

    iph->ip_len = ntohs(iph->ip_len);
    iph->ip_id  = ntohs(iph->ip_id);
    iph->ip_off = ntohs(iph->ip_off);

    if ((iph->ip_off & IP_MF) || (iph->ip_off & IP_OFFMASK))
	return;

    datalen = iph->ip_len - (iph->ip_hl << 2) - sizeof(icmph_t);
    icmph   = (icmph_t *)(p + (iph->ip_hl << 2));

    if (tstamp) {
         gettimeofday(&tp, NULL);
         tm = localtime((long *)&tp.tv_sec);
         printf("%02d%02d%04d.%02d%02d.%02d ", tm->tm_mday, tm->tm_mon, 
		1900 + tm->tm_year, tm->tm_hour, tm->tm_min, tm->tm_sec);
    }

    memset(ip_dname, '\0', sizeof(ip_dname));

    if (hostnames) {
	ia.s_addr = iph->ip_src;
	if ((hps = gethostbyaddr((char *)&ia, 4, AF_INET)) != NULL)
	    snprintf(ip_sname, sizeof(ip_sname), "%s", hps->h_name);

	if (promisc) {
	    ia.s_addr = iph->ip_dst;
	    if ((hpd = gethostbyaddr((char *)&ia, 4, AF_INET)) != NULL)
	        snprintf(ip_dname, sizeof(ip_dname), "%s", hpd->h_name);
	}

    }       

   /*
    * Since we could be pinging from a multi-homed host we get the
    * source host everytime.
    */
    printf("Ping from %s ", ip_sname[0] != '\0' ? ip_sname : 
		inet_nbotoa(iph->ip_src));
    if (promisc)
        printf("=> %s ", ip_dname[0] != '\0' ? ip_dname : inet_nbotoa(iph->ip_dst));
    printf("is ");

    for ( ; e->signature != NULL; e++) {

	if (datalen != (e->length + e->offset))
	    continue;

	dataoffset = (char *)(p + (iph->ip_hl << 2) + 
			sizeof(icmph_t) + e->offset);

	if (memcmp(e->signature, dataoffset, e->length) != 0)
	    continue;

	identified++;

	for (namep = e->osnames; *namep != NULL; namep++)
            printf("%s ", *namep);
    }

    if (identified == 0)
	printf("Unknown\n");
    else if (identified > 1) {
	printf("\nDatabase Problems\n");
	exit(1);
    } else
	printf("\n"); 

    if (verbose > 1) 
        print_echo((char *)icmph, datalen);

    return;
}

void usage(void) {
    printf("blah: [-v[v]pnth?] [-i interface]\n");
}

int main(int argc, char **argv)
{
    pcap_t *pd;
    bpf_u_int32 netp;
    bpf_u_int32 maskp;
    struct bpf_program fp;
    int    i = 0;
    char   errbuf[PCAP_ERRBUF_SIZE];
    char   *filter    = NULL;
    char   *interface = NULL;

    while ((i = getopt(argc, argv, "ni:pvht?")) != EOF) {
        switch (i) {
	    case 'n':
		hostnames = 0;
		break;
            case 'i': 
		interface = optarg;
                break;
            case 'p': 
		promisc  = 1;
		break;
            case 'v': 
		verbose++;
		break;
            case 't':
                tstamp = 0;
                break;
            case 'h':
            case '?':
	    default:
		usage();
		return -1;
       }
    }

    if (interface == NULL && ((interface = pcap_lookupdev(errbuf)) == NULL)) {
	printf("pcap_lookupdev(): %s\n", errbuf);
	return -1;
    }

    if ((pd = pcap_open_live(interface, 1500, promisc, 20, errbuf)) == NULL) {
	printf("pcap_open_live(): %s\n", errbuf);
	return -1;
    }

    filter = "ip[0:1]=0x45 and ip[2:2] >= 28 and ip[2:2] <= 1500 and icmp[0:1] = 8";

    if (pcap_lookupnet(interface, &netp, &maskp, errbuf) < 0) {
	printf("pcap_lookupnet(): %s\n", errbuf);
	return -1;
    }

    if (pcap_compile(pd, &fp, filter, 1, maskp) < 0) {
	printf("pcap_compile()\n");
	return -1;
    }

    if (pcap_setfilter(pd, &fp) < 0) {
	printf("pcap_setfilter()\n");
	return -1;
    }

    switch(linklayer = pcap_datalink(pd)) {
	case DLT_EN10MB:
	    linklayer = 14;
	    break;
	case DLT_NULL:
	    linklayer = 4;
	    break;
	default:
	    printf("pcap_datalink(): Unknown linklayer\n");
	    return -1;
    }

    memset(ip_sname, '\0', sizeof(ip_sname));
    gethostname(ip_sname, sizeof(ip_sname));
    printf("Sniffing on host %s interface %s\n", ip_sname, interface);
    memset(ip_sname, '\0', sizeof(ip_sname));
    pcap_loop(pd, -1, ping_ident, NULL);
    return -1;

}
